package aggSim.analyzer;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.gef.EditPart;
import org.eclipse.gef.ui.actions.SelectionAction;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IWorkbenchPart;

import roneditor.model.ObjectNetUtil;
import ronmodel.ObjectNet;
import ronmodel.Rule;
import aggSim.Grammar;
import aggSim.analyzer.dialogs.RuleRuleConflictDialog;
import aggSim.analyzer.dialogs.RuleTransConflictDialog;
import aggSim.analyzer.dialogs.TransTransConflictDialog;

/**
 * {@link SelectionAction} for performing an conflict analysis on either an
 * {@link ObjectNet} (Transition/Transition conflict), an {@link ObjectNet} and
 * a {@link Rule} (Transition/Rule application conflict) or an {@link ObjectNet}
 * and two {@link Rule}s (Rule application/Rule application conflict). These may
 * be freely selected in the editor. If more than one object net is selected the
 * result of their union is considered for analysis. The result of the analysis
 * is shown in an appropriate dialog.
 * 
 * Analysis is performed on copies of the actual models! So the dialogs showing
 * the results could not harm the working model.
 * 
 * @author Tony Modica
 */
public class ConflictAnalysisAction extends SelectionAction {

	public static final String ID = "roneditor.actions.ConflictAnalysisAction";

	public static final String NAME = "Analyze conflicts";

	private static final ArrayList<Rule> selectedRules = new ArrayList<Rule>();

	private static final ArrayList<ObjectNet> selectedNets = new ArrayList<ObjectNet>();

	public ConflictAnalysisAction(final IWorkbenchPart part) {
		super(part);
		setId(ID);
		setText(NAME);
	}

	@Override
	public void run() {
		final Grammar grammar = new Grammar();

		// this is already a "copy" of the selected object nets!
		final ObjectNet selectedObjectNet = ObjectNetUtil
				.unionIntoObjectNet(selectedNets);
		List<Conflict> conflicts;
		String title = "Error";
		String message = "Some error occured during conflict computation!";
		switch (selectedRules.size()) {
		case 0:
			String conflictKind = "Transition/Transition conflicts";
			Dialog waitDialog = openWaitingDialog(conflictKind);
			conflicts = Analyzer.analyzeTokenFiring(selectedObjectNet);
			waitDialog.close();
			if (!conflicts.isEmpty()) {
				final TransTransConflictDialog diag = new TransTransConflictDialog(
						selectedObjectNet, conflicts);
				diag.setDoLayout(selectedObjectNet.eContainer() == null);
				diag.open();
				return;
			}
			title = "No " + conflictKind + " found";
			message = "There are no conflicts in the union of the selected object nets!";
			break;
		case 1:
			conflictKind = "Rule/Transition conflicts";
			// this is already a copy of the selected rule!
			final Rule selectedRule = selectedRules.get(0);
			if (!grammar.isApplicable(selectedRule, selectedObjectNet)) {
				title = "No " + conflictKind + " found";
				message = "There are no conflicts as the rule '"
						+ selectedRule.getName()
						+ "' is not applicable to the union of the selected object nets!";
				break;
			}
			waitDialog = openWaitingDialog(conflictKind);
			conflicts = Analyzer.analyzeFiringTransformation(selectedRule,
					selectedObjectNet);
			waitDialog.close();
			if (!conflicts.isEmpty()) {
				final RuleTransConflictDialog diag = new RuleTransConflictDialog(
						selectedObjectNet, selectedRule, conflicts);
				diag.setDoLayout(selectedObjectNet.eContainer() == null);
				diag.open();
				return;
			}
			title = "No " + conflictKind + " found";
			message = "There are no conflicts between the rule '"
					+ selectedRule.getName()
					+ "' and the union of the selected object nets!";
			break;
		case 2:
			conflictKind = "Rule/Rule conflicts";
			// these are already copies of the selected rules!
			final Rule firstRule = selectedRules.get(0);
			final Rule secondRule = selectedRules.get(1);
			if (!grammar.isApplicable(firstRule, selectedObjectNet)) {
				title = "No possible " + conflictKind;
				message = "The rule '"
						+ firstRule.getName()
						+ "' is not applicable to the union of the selected object nets! So there are no conflicts to rule '"
						+ secondRule.getName() + "'.";
				break;
			}
			if (!grammar.isApplicable(secondRule, selectedObjectNet)) {
				title = "No possible " + conflictKind;
				message = "The rule '"
						+ secondRule.getName()
						+ "' is not applicable to the union of the selected object nets! So there are no conflicts to rule '"
						+ firstRule.getName() + "'.";
				break;
			}
			waitDialog = openWaitingDialog(conflictKind);
			conflicts = Analyzer.analyzeNetTransformation(selectedObjectNet,
					firstRule, secondRule);
			waitDialog.close();
			if (!conflicts.isEmpty()) {
				final RuleRuleConflictDialog diag = new RuleRuleConflictDialog(
						selectedObjectNet, firstRule, secondRule, conflicts);
				diag.setDoLayout(selectedObjectNet.eContainer() == null);
				diag.open();
				return;
			}
			title = "No " + conflictKind + " found";
			message = "There are no conflicts between the rule '"
					+ firstRule.getName()
					+ "' and the rule '"
					+ secondRule.getName()
					+ "' when applied to the union of the selected object nets!";
			break;
		default:
			break;
		}

		// if we reach this point no dialog has been shown before, so show
		// actual message and title
		MessageDialog.openInformation(null, title, message);
	}

	private Dialog openWaitingDialog(final String conflictKind) {
		final Dialog dialog = new MessageDialog(Display.getDefault()
				.getActiveShell(), "Computing " + conflictKind, null,
				"Please wait for the " + conflictKind
						+ " to be computed.\nThis is not cancelable!",
				MessageDialog.INFORMATION, new String[] {}, 0) {
			@Override
			protected Control createButtonBar(final Composite parent) {
				return null;
			}
		};
		dialog.setBlockOnOpen(false);
		dialog.open();
		return dialog;
	}

	/**
	 * This {@link SelectionAction} is enabled if the selection meets the
	 * requirements of one analysis method described in the class description.
	 */
	@Override
	protected boolean calculateEnabled() {
		setText(NAME);
		// presort selected elements to check analysis type later in run()
		selectedRules.clear();
		selectedNets.clear();
		for (final Object selectedObject : getSelectedObjects()) {
			if (selectedObject instanceof EditPart) {
				final Object selectedModel = ((EditPart) selectedObject)
						.getModel();
				if (selectedModel instanceof ObjectNet) {
					selectedNets.add((ObjectNet) selectedModel);
				}
				if (selectedModel instanceof Rule) {
					selectedRules.add((Rule) selectedModel);
				}
			}
		}

		// has at least one ObjectNet and not more than two rules been selected?
		if (selectedNets.isEmpty() || selectedRules.size() > 2) {
			return false;
		}

		// make sure nothing else has been selected (as it may be confusing for
		// the user, if this action is enabled)
		if (getSelectedObjects().size() != selectedRules.size()
				+ selectedNets.size()) {
			return false;
		}

		// update action name
		switch (selectedRules.size()) {
		case 0:
			setText(NAME + " (T/T)");
			break;
		case 1:
			setText(NAME + " (R/T)");
			break;
		case 2:
			setText(NAME + " (R/R)");
			break;
		}
		return true;
	}
}
