package roneditor;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import muvitorkit.actions.DNPTBAction;
import muvitorkit.actions.GenericCopyAction;
import muvitorkit.actions.GenericCutAction;
import muvitorkit.actions.GenericPasteAction;
import muvitorkit.ui.ContextMenuProviderWithActionRegistry;
import muvitorkit.ui.MuvitorTreeEditor;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.gef.EditPartFactory;
import org.eclipse.gef.KeyHandler;
import org.eclipse.gef.ui.parts.TreeViewer;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorSite;

import roneditor.gef.actions.ron.CreateHLPlaceAction;
import roneditor.gef.actions.ron.CreateTokenAction;
import roneditor.gef.actions.ron.RunAutoSimulationAction;
import roneditor.gef.actions.ron.StopAutoSimulationAction;
import roneditor.gef.editparts.tree.RONTreeEditPartFactory;
import roneditor.model.HLTransitionUtil;
import roneditor.ui.ron.RONEditorContextMenuProvider;
import ronmodel.HLPlace;
import ronmodel.HLPlaceType;
import ronmodel.Mapping;
import ronmodel.Node;
import ronmodel.ONNode;
import ronmodel.ObjectNet;
import ronmodel.RON;
import ronmodel.RonmodelFactory;
import ronmodel.RonmodelPackage;
import ronmodel.Rule;

/**
 * This class extends {@link MuvitorTreeEditor} to an editor for Reconfigurable
 * Object Nets (RONs).
 * 
 * @author Tony Modica
 */
public class RONTreeEditor extends MuvitorTreeEditor {

	/**
	 * This is the ID of the View for object nets according to
	 * <code>plugin.xml</code>.
	 */
	public final static String objectNetViewID = "roneditor.ui.views.ObjectNetPageBookView";

	/**
	 * This is the ID of the View for rules according to <code>plugin.xml</code>
	 * .
	 */
	public final static String ruleViewID = "roneditor.ui.views.RulePageBookView";

	/**
	 * This is the ID of the View for the ron according to
	 * <code>plugin.xml</code>.
	 */
	public final static String ronViewID = "roneditor.ui.views.RONPageBookView";

	// register the viewIDs once statically
	static {
		MuvitorTreeEditor.registerViewID(RonmodelPackage.Literals.RON,
				ronViewID);
		MuvitorTreeEditor.registerViewID(RonmodelPackage.Literals.RULE,
				ruleViewID);
		MuvitorTreeEditor.registerViewID(RonmodelPackage.Literals.OBJECT_NET,
				objectNetViewID);
	}

	public RONTreeEditor() {
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see muvitorkit.ui.AbstractTreeEditor#createDefaultModel()
	 */
	@Override
	protected EObject createDefaultModel() {
		final RON newRon = RonmodelFactory.eINSTANCE.createRON();
		newRon.setName("<default>");
		return newRon;
	}

	/**
	 * Overridden to show the loaded RON
	 */
	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * roneditor.ui.part.AbstractTreeEditor#init(org.eclipse.ui.IEditorSite,
	 * org.eclipse.ui.IEditorInput)
	 */
	@Override
	public void init(final IEditorSite site, final IEditorInput input) {
		super.init(site, input);
		// show RON, if available after initializing
		if (getPrimaryModelRoot() != null && getSite().getPage() != null) {
			// this wrapping is needed because the job is not in the UI
			// thread
			Display.getDefault().asyncExec(new Runnable() {
				@Override
				public void run() {
					HLTransitionUtil.invalidate((RON) getPrimaryModelRoot());
					showView(getPrimaryModelRoot());
				}
			});
		}
	}

	/**
	 * Show the RON after reverting.
	 * 
	 * @see muvitorkit.ui.MuvitorTreeEditor#revertToLastSaved()
	 */
	@Override
	public void revertToLastSaved() {
		super.revertToLastSaved();
		if (getPrimaryModelRoot() != null && getSite().getPage() != null) {
			HLTransitionUtil.invalidate((RON) getPrimaryModelRoot());
			showView(getPrimaryModelRoot());
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see roneditor.editor.AbstractTreeEditor#getTreeEditPartFactory()
	 */
	@Override
	protected EditPartFactory createTreeEditPartFactory() {
		return new RONTreeEditPartFactory();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * roneditor.editor.AbstractTreeEditor#getContextMenuProvider(org.eclipse
	 * .gef.ui.parts.TreeViewer)
	 */
	@Override
	protected ContextMenuProviderWithActionRegistry createContextMenuProvider(
			final TreeViewer viewer) {
		return new RONEditorContextMenuProvider(viewer);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see roneditor.ui.part.AbstractTreeEditor#createCustomActions()
	 */
	@Override
	protected void createCustomActions() {
		registerAction(new GenericCopyAction(this));
		registerAction(new GenericCutAction(this));

		// define rules for paste actions
		final GenericPasteAction pasteAction = new GenericPasteAction(this);
		pasteAction.registerPasteRule(RonmodelPackage.Literals.TOKEN,
				new GenericPasteAction.PasteRuleAdapter() {
					@Override
					public boolean isValidPaste(final EObject element,
							final EObject target) {
						final HLPlaceType placeType = ((HLPlace) target)
								.getType();
						if (element instanceof ObjectNet
								&& placeType != HLPlaceType.NET_LITERAL) {
							return false;
						} else if (element instanceof Rule
								&& placeType != HLPlaceType.RULE_LITERAL) {
							return false;
						}
						return true;
					}
				});

		pasteAction.registerPasteRule(RonmodelPackage.Literals.NODE,
				new GenericPasteAction.PasteRuleAdapter() {
					@Override
					public void afterPaste(final EObject element,
							final EObject target) {
						final Node node = (Node) element;
						node.setX(node.getX() + 100);
						node.setY(node.getY() + 100);
					}
				});

		registerAction(pasteAction);

		registerAction(new CreateHLPlaceAction(this));

		registerAction(new CreateTokenAction(this));

		registerActionOnToolBar(new RunAutoSimulationAction(this))
				.addPropertyChangeListener(new IPropertyChangeListener() {
					@SuppressWarnings("synthetic-access")
					@Override
					public void propertyChange(final PropertyChangeEvent event) {
						updateActions();
					}
				});
		registerActionOnToolBar(new StopAutoSimulationAction(this));
		registerActionOnToolBar(new DNPTBAction());

		// registerAction(new ExportPNMLAction(this));
	}

	/**
	 * FIXED: Need to remove match mappings (which have no parent) to be able to
	 * save the model!
	 */
	@Override
	protected void save(final IFile file, final IProgressMonitor monitor)
			throws CoreException {
		final Map<Mapping, ONNode> mapping2Source = new HashMap<Mapping, ONNode>();
		final Map<Mapping, ONNode> mapping2Target = new HashMap<Mapping, ONNode>();
		removeUnparentedMappings(getPrimaryModelRoot(), mapping2Source,
				mapping2Target);
		super.save(file, monitor);
		for (final Entry<Mapping, ONNode> entry : mapping2Source.entrySet()) {
			entry.getKey().setSource(entry.getValue());
		}
		for (final Entry<Mapping, ONNode> entry : mapping2Target.entrySet()) {
			entry.getKey().setTarget(entry.getValue());
		}
	}

	private static void removeUnparentedMappings(final EObject model,
			final Map<Mapping, ONNode> mapping2Source,
			final Map<Mapping, ONNode> mapping2Target) {
		for (final EObject child : model.eContents()) {
			if (child instanceof ONNode) {
				final ONNode node = (ONNode) child;
				for (final Mapping mapping : new ArrayList<Mapping>(
						node.getIsSourceInMappings())) {
					if (mapping.eResource() == null) {
						mapping2Source.put(mapping, node);
						mapping.setSource(null);
					}
				}
				for (final Mapping mapping : new ArrayList<Mapping>(
						node.getIsTargetInMappings())) {
					if (mapping.eResource() == null) {
						mapping2Target.put(mapping, node);
						mapping.setTarget(null);
					}
				}
			} else {
				removeUnparentedMappings(child, mapping2Source, mapping2Target);
			}
		}
	}

	@Override
	protected void setupKeyHandler(final KeyHandler kh) {
	}
}
