package roneditor.gef.commands.objectnet;

import java.util.ArrayList;

import muvitorkit.animation.AnimatingCommand;
import muvitorkit.animation.AnimationPathModifier;

import org.eclipse.draw2d.ColorConstants;
import org.eclipse.draw2d.Ellipse;
import org.eclipse.draw2d.IFigure;
import org.eclipse.gef.commands.Command;
import org.eclipse.gef.commands.CompoundCommand;

import roneditor.model.ONTransitionUtil;
import ronmodel.ONArc;
import ronmodel.ONPlace;
import ronmodel.ONTransition;
import ronmodel.ObjectNet;

/**
 * Command to fire a transition in a petrinet.
 */
public class FireTransitionCommand extends CompoundCommand {

	private static final String LABEL = "Fire Transition '";

	public FireTransitionCommand(final ONTransition transition) {
		super(LABEL + transition.getName() + "' in '"
				+ transition.getObjectNet().getName() + "'");

		// animate low-level tokens to transition
		AnimatingCommand aniComm = new AnimatingCommand();
		aniComm.setCompleteDuration(1000);
		final ObjectNet objectNet = transition.getObjectNet();
		final ArrayList<IFigure> tokenFigures = new ArrayList<IFigure>();
		for (final ONArc arc : transition.getIncoming()) {
			final Ellipse tokenFigure = createTokenFigure();
			final AnimationPathModifier pm = AnimationPathModifier
					.getSineModifier(20, 2);
			aniComm.initializeAnimatedElement(tokenFigure, objectNet, pm);
			aniComm.specifyStep(tokenFigure, arc.getSource());
			// remember animated figures for next step
			tokenFigures.add(tokenFigure);
		}
		aniComm.nextStep();
		for (final IFigure tokenFigure : tokenFigures) {
			// each incoming arc gets its token
			aniComm.specifyStep(tokenFigure, transition);
		}
		add(aniComm);

		// animate low-level tokens from transition
		aniComm = new AnimatingCommand();
		aniComm.setCompleteDuration(1000);
		tokenFigures.clear();
		for (int i = 0; i < transition.getOutgoing().size(); i++) {
			final Ellipse tokenFigure = createTokenFigure();
			aniComm.initializeAnimatedElement(tokenFigure, objectNet);
			aniComm.specifyStep(tokenFigure, transition);
			// assign the animated figures to the outgoing arcs
			tokenFigures.add(tokenFigure);
		}
		aniComm.nextStep();
		int tokenCounter = 0;
		for (final ONArc arc : transition.getOutgoing()) {
			// remember animated figures for next step
			aniComm.specifyStep(tokenFigures.get(tokenCounter++),
					arc.getTarget());
		}
		add(aniComm);

		// a command doing the actual work
		final Command command = new Command() {
			/**
			 * Executes the command by decreasing/increasing the number of
			 * tokens first places according second the weight of the arc
			 * connecting a place with the transition being fired.
			 * 
			 * @see org.eclipse.gef.commands.Command#execute()
			 */
			@Override
			public void execute() {
				// Decrease tokens of all source places
				for (final ONArc incomingArc : transition.getIncoming()) {
					final ONPlace sourcePlace = (ONPlace) incomingArc
							.getSource();
					sourcePlace.setTokens(sourcePlace.getTokens()
							- incomingArc.getWeight());
				}

				// Increase tokens of all target places
				for (final ONArc outgoingArc : transition.getOutgoing()) {
					final ONPlace targetPlace = (ONPlace) outgoingArc
							.getTarget();
					targetPlace.setTokens(targetPlace.getTokens()
							+ outgoingArc.getWeight());
				}
			}

			@Override
			public boolean canExecute() {
				return ONTransitionUtil.canFire(transition);
			}

			@Override
			public void undo() {
				// Decrease tokens of all source places
				for (final ONArc incomingArc : transition.getIncoming()) {
					final ONPlace sourcePlace = (ONPlace) incomingArc
							.getSource();
					sourcePlace.setTokens(sourcePlace.getTokens()
							+ incomingArc.getWeight());
				}

				// Increase tokens of all target places
				for (final ONArc outgoingArc : transition.getOutgoing()) {
					final ONPlace targetPlace = (ONPlace) outgoingArc
							.getTarget();
					targetPlace.setTokens(targetPlace.getTokens()
							- outgoingArc.getWeight());
				}

			}
		};
		add(command);
	}

	private Ellipse createTokenFigure() {
		final Ellipse tokenFigure = new Ellipse();
		tokenFigure.setSize(10, 10);
		tokenFigure.setBackgroundColor(ColorConstants.black);
		return tokenFigure;
	}
}
