package roneditor.gef.commands.ron;

import java.util.Collection;

import muvitorkit.animation.AnimatingCommand;

import org.eclipse.emf.common.util.EList;
import org.eclipse.gef.commands.Command;
import org.eclipse.gef.commands.CompoundCommand;

import roneditor.model.HLTransitionUtil;
import roneditor.model.ObjectNetUtil;
import ronmodel.HLPlace;
import ronmodel.HLTransition;
import ronmodel.ONNode;
import ronmodel.ObjectNet;
import ronmodel.RON;
import ronmodel.Token;

/**
 * The {@link HLTransitionSplitCommand} is used when we fire an
 * {@link HLTransition} of type <code>SPLIT</code>. We can only execute it when
 * there is one incoming {@link HLPlace} of type <code>NET</code> and one
 * outgoing {@link HLPlace} of type <code>NET</code>. The
 * {@link CompoundCommand} will then split the {@link Token} (which are in turn
 * an {@link ObjectNet}) placed on the source {@link HLPlace} in it's connected
 * component and put the generated {@link ObjectNet}s (which are in turn
 * {@link Token}s again) on the target {@link HLPlace}.
 */
public class HLTransitionSplitCommand extends CompoundCommand {

	private static final String LABEL = "Split object net '";

	private static class SplitObjectNetCommand extends Command {

		private final HLPlace sourcePlace;

		private final HLPlace targetPlace;

		private final ObjectNet sourceToken;

		private Collection<ObjectNet> targetToken;

		SplitObjectNetCommand(final ObjectNet token, final HLPlace target) {
			sourceToken = token;
			targetPlace = target;
			sourcePlace = token.getPlace();
		}

		@Override
		public void execute() {
			targetToken = ObjectNetUtil.split(sourceToken);
			for (final ObjectNet net : targetToken) {
				net.setPlace(targetPlace);
			}
			sourceToken.setPlace(null);
		}

		@Override
		public void undo() {
			sourceToken.setPlace(sourcePlace);
			for (final ObjectNet net : targetToken) {
				final Collection<ONNode> nodes = net.getNodes();
				ObjectNetUtil.moveConnectedComponent(nodes, sourceToken);
				net.setPlace(null);
			}
		}
	}

	private final HLTransition transition;

	public HLTransitionSplitCommand(final HLTransition transition) {
		this.transition = transition;
		// Get connected HLPlaces
		final HLPlace sourcePlace = (HLPlace) transition.getIncoming().get(0)
				.getSource();
		final HLPlace targetPlace = (HLPlace) transition.getOutgoing().get(0)
				.getTarget();

		final ObjectNet token = (ObjectNet) sourcePlace.getTokens().get(0);

		/*
		 * muvitorkit.animation: move objectnets from pre-domain places to
		 * transition, done without overwriting initialize() because at this
		 * time all elements to be animated exist (but could have done so ntl.)
		 */
		final RON ron = transition.getRon();
		final AnimatingCommand aniComm = new AnimatingCommand();
		aniComm.setCompleteDuration(1000);
		aniComm.initializeAnimatedElement(token, ron);
		aniComm.specifyStep(token, token.getPlace());
		aniComm.nextStep();
		aniComm.specifyStep(token, transition);
		add(aniComm);

		add(new SplitObjectNetCommand(token, targetPlace));

		/*
		 * animation: move rule and new unionObjectNets from transition to
		 * post-domain places, done by overwriting initialize because elements
		 * to be animated will not exist before compound command is being
		 * executed
		 */

		// remember old token number on target place to determine the new
		// added tokens to be animated
		final int oldTokenNumber = targetPlace.getTokens().size();

		final AnimatingCommand aniComm2 = new AnimatingCommand() {
			@Override
			public void initialize() {
				setCompleteDuration(1000);
				final EList<Token> tokens = targetPlace.getTokens();
				// the last tokens on outgoingPlaces should be the tokens to
				// be animated
				for (int i = oldTokenNumber; i < tokens.size(); i++) {
					final Token newToken = tokens.get(i);
					initializeAnimatedElement(newToken, ron);
					specifyStep(newToken, transition);
				}
				nextStep();
				for (int i = oldTokenNumber; i < tokens.size(); i++) {
					final Token newToken = tokens.get(i);
					specifyStep(newToken, targetPlace);
				}

			}
		};
		add(aniComm2);

		setLabel(LABEL + token.getName() + "' on HLTrans '"
				+ transition.getName() + "'");
	}

	@Override
	public void execute() {
		// before animating close views showing object nets that have been
		// removed
		HLTransitionUtil.closeViewsForConsumedObjectNets(transition);
		super.execute();
	}

}
