package roneditor.gef.commands.ron;

import java.util.ArrayList;
import java.util.List;

import muvitorkit.animation.AnimatingCommand;

import org.eclipse.emf.common.util.EList;
import org.eclipse.gef.commands.CompoundCommand;

import roneditor.model.HLTransitionUtil;
import roneditor.model.ObjectNetUtil;
import ronmodel.HLPlace;
import ronmodel.HLPlaceType;
import ronmodel.HLTransition;
import ronmodel.ObjectNet;
import ronmodel.RON;
import ronmodel.Token;

/**
 * Perform <code>STANDARD</code> operation on the {@link HLTransition}.
 * <ul>
 * <li>We copy the contents of <i>all</i> {@link Token} on incoming
 * {@link HLPlace}s from the {@link HLPlaceType#NET} into the same number of
 * {@link ObjectNet}s as we have outgoing {@link HLPlace} of the
 * {@link HLPlaceType#NET}, put the new {@link Token}s onto the targets and
 * remove the original {@link Token}s from the source.</li>
 * 
 * <li>As we require to have at most <i>one</i> {@link HLPlace} of
 * {@link HLPlaceType#RULE}, we just <i>move</i> the {@link Token} to the
 * target.</li>
 * </ul>
 */
public class HLTransitionStandardCommand extends CompoundCommand {

	private static final String LABEL = "Fire HLTrans ";
	private final HLTransition transition;

	public HLTransitionStandardCommand(final HLTransition transition) {
		super(LABEL + "'" + transition.getName() + "'");
		this.transition = transition;
		// Collect incoming NET/RULE HLPlaces
		final List<HLPlace> incomingNetPlaces = HLTransitionUtil
				.getIncomingNetPlaces(transition);
		final List<HLPlace> incomingRulePlaces = HLTransitionUtil
				.getIncomingRulePlaces(transition);

		// Collect outgoing NET/RULE HLPlaces
		final List<HLPlace> outgoingNetPlaces = HLTransitionUtil
				.getOutgoingNetPlaces(transition);
		final List<HLPlace> outgoingRulePlaces = HLTransitionUtil
				.getOutgoingRulePlaces(transition);

		final List<ObjectNet> netTokens = HLTransitionUtil
				.collectSourceNetTokens(transition);

		Token ruleToken = null;
		// process RulePlaces
		if (!incomingRulePlaces.isEmpty()) {
			final HLPlace incomingRulePlace = incomingRulePlaces.get(0);
			ruleToken = incomingRulePlace.getTokens().get(0);
		}

		/*
		 * animation: move objectnets and rules from pre-domain places to
		 * transition, done without overwriting initialize() because at this
		 * time all elements to be animated exist (but could have done so ntl.)
		 */
		final RON ron = transition.getRon();
		final AnimatingCommand aniComm = new AnimatingCommand();
		aniComm.setCompleteDuration(1000);
		for (final ObjectNet token : netTokens) {
			aniComm.initializeAnimatedElement(token, ron);
			aniComm.specifyStep(token, token.getPlace());
		}
		if (ruleToken != null) {
			aniComm.initializeAnimatedElement(ruleToken, ron);
			aniComm.specifyStep(ruleToken, ruleToken.getPlace());
		}
		aniComm.nextStep();
		for (final ObjectNet token : netTokens) {
			aniComm.specifyStep(token, transition);
		}
		if (ruleToken != null) {
			aniComm.specifyStep(ruleToken, transition);
		}
		add(aniComm);

		if (!incomingRulePlaces.isEmpty()) {
			// Copy rule to the new outgoing tokens on HLPlace.RULE
			for (final HLPlace place : outgoingRulePlaces) {
				add(new CopyTokenToHLPlaceCommand(ruleToken, place, false));
			}
		}

		if (!incomingNetPlaces.isEmpty()) {
			// union of all first token of incomingNetPlaces into one
			// unionObjectNet
			final ObjectNet unionObjectNet = ObjectNetUtil
					.unionIntoObjectNet(netTokens);
			// Copy unionObjectNet to the new outgoing tokens on HLPlace.NET
			for (final HLPlace place : outgoingNetPlaces) {
				add(new CopyTokenToHLPlaceCommand(unionObjectNet, place, false));
			}
		}

		// remove tokens/ObjectNets from place
		for (final ObjectNet t : netTokens) {
			add(new DeleteTokenCommand(t));
		}
		if (ruleToken != null) {
			// remove rule from the HLPlace of Type RULE
			add(new DeleteTokenCommand(ruleToken));
		}

		/*
		 * animation: move rule and new unionObjectNets from transition to
		 * post-domain places, done by overwriting initialize because elements
		 * to be animated will not exist before compound command is being
		 * executed
		 */
		final AnimatingCommand aniComm2 = new AnimatingCommand() {
			@Override
			public void initialize() {
				setCompleteDuration(1000);
				// the last tokens on outgoingPlaces should be the tokens to
				// be animated
				final List<HLPlace> outgoingPlaces = new ArrayList<HLPlace>();
				outgoingPlaces.addAll(outgoingNetPlaces);
				outgoingPlaces.addAll(outgoingRulePlaces);

				for (final HLPlace place : outgoingPlaces) {
					final EList<Token> tokens = place.getTokens();
					final Token token = tokens.get(tokens.size() - 1);
					initializeAnimatedElement(token, ron);
					specifyStep(token, transition);
				}
				nextStep();
				for (final HLPlace place : outgoingPlaces) {
					final EList<Token> tokens = place.getTokens();
					final Token token = tokens.get(tokens.size() - 1);
					specifyStep(token, place);
				}
			}
		};
		add(aniComm2);
	}

	@Override
	public void execute() {
		// before animating close views showing object nets that have been
		// removed
		HLTransitionUtil.closeViewsForConsumedObjectNets(transition);
		super.execute();
	}

}
