package roneditor.gef.editparts.objectnet;

import muvitorkit.gef.directedit.IDirectEditPart.IGraphicalDirectEditPart;
import muvitorkit.gef.editparts.AdapterGraphicalEditPart;
import muvitorkit.ui.utils.SWTResourceManager;

import org.eclipse.core.runtime.Assert;
import org.eclipse.draw2d.ConnectionAnchor;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.util.EList;
import org.eclipse.gef.ConnectionEditPart;
import org.eclipse.gef.EditPolicy;
import org.eclipse.gef.GraphicalEditPart;
import org.eclipse.gef.NodeEditPart;
import org.eclipse.gef.Request;
import org.eclipse.jface.viewers.ICellEditorValidator;
import org.eclipse.swt.graphics.RGB;

import roneditor.gef.editparts.objectnet.policies.ONNodeComponentEditPolicy;
import roneditor.gef.editparts.objectnet.policies.ONPlaceEditPolicy;
import roneditor.gef.figures.objectnet.NodeFigure;
import ronmodel.NodeColor;
import ronmodel.ONArc;
import ronmodel.ONNode;
import ronmodel.RonmodelPackage;

/**
 * EditPart for a generic ONNode. The derived {@link ONTransitionEditPart} and
 * {@link ONPlaceEditPolicy} need to implement the deviant functionality.
 */
abstract public class ONNodeEditPart<T extends ONNode> extends
		AdapterGraphicalEditPart<T> implements NodeEditPart,
		IGraphicalDirectEditPart {

	/**
	 * Creates a new {@link ONNodeEditPart} instance.
	 * 
	 * @param node
	 *            The Node
	 */
	protected ONNodeEditPart(final T node) {
		super(node);
	}

	/**
	 * Get the incoming connections of the node
	 * 
	 * @return List of ONNodes
	 * @see org.eclipse.gef.editparts.AbstractGraphicalEditPart#getModelSourceConnections()
	 */
	@Override
	protected EList<ONArc> getModelSourceConnections() {
		return getCastedModel().getOutgoing();
	}

	/**
	 * Get the incoming connections of the node
	 * 
	 * @return List of ONNodes
	 * @see org.eclipse.gef.editparts.AbstractGraphicalEditPart#getModelTargetConnections()
	 */
	@Override
	protected EList<ONArc> getModelTargetConnections() {
		return getCastedModel().getIncoming();
	}

	/**
	 * Returns the Figure of this, as a transition type figure.
	 * 
	 * @return Figure as a TransitionFigure.
	 */
	protected NodeFigure getNodeFigure() {
		return (NodeFigure) getFigure();
	}

	/**
	 * Create all basic policies for a ONNode.
	 * 
	 * @see org.eclipse.gef.editparts.AbstractEditPart#createEditPolicies()
	 */
	@Override
	protected void createEditPolicies() {
		// install the edit policy to handle deletion
		installEditPolicy(EditPolicy.COMPONENT_ROLE,
				new ONNodeComponentEditPolicy());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.gef.NodeEditPart#getSourceConnectionAnchor(org.eclipse.gef
	 * .ConnectionEditPart)
	 */
	@Override
	public ConnectionAnchor getSourceConnectionAnchor(
			final ConnectionEditPart connection) {
		return getNodeFigure().getSourceConnectionAnchor();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.gef.NodeEditPart#getSourceConnectionAnchor(org.eclipse.gef
	 * .Request)
	 */
	@Override
	public ConnectionAnchor getSourceConnectionAnchor(final Request request) {
		return getNodeFigure().getSourceConnectionAnchor();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.gef.NodeEditPart#getTargetConnectionAnchor(org.eclipse.gef
	 * .ConnectionEditPart)
	 */
	@Override
	public ConnectionAnchor getTargetConnectionAnchor(
			final ConnectionEditPart connection) {
		return getNodeFigure().getTargetConnectionAnchor();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.gef.NodeEditPart#getTargetConnectionAnchor(org.eclipse.gef
	 * .Request)
	 */
	@Override
	public ConnectionAnchor getTargetConnectionAnchor(final Request request) {
		return getNodeFigure().getTargetConnectionAnchor();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.gef.editparts.AbstractEditPart#refreshVisuals()
	 */
	@Override
	protected void refreshVisuals() {
		// get rid of old colors
		final RGB figureRGB = getNodeFigure().getColor().getRGB();
		final NodeColor modelColor = getCastedModel().getColor();
		if (modelColor != null) {
			final RGB modelRGB = new RGB(modelColor.getRed(),
					modelColor.getGreen(), modelColor.getBlue());
			if (!figureRGB.equals(modelRGB)) {
				getNodeFigure().setColor(SWTResourceManager.getColor(modelRGB));
			}
		}
		getNodeFigure().updateValueDisplay(getCastedModel().getName());
		final Point loc = new Point(getCastedModel().getX(), getCastedModel()
				.getY());
		final Dimension size = new Dimension(-1, -1);
		final Rectangle r = new Rectangle(loc, size);
		((GraphicalEditPart) getParent()).setLayoutConstraint(this,
				getFigure(), r);
		getNodeFigure().validate();
		super.refreshVisuals();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * muvitorkit.gef.editparts.AdapterGraphicalEditPart#notifyChanged(org.eclipse
	 * .emf.common.notify.Notification)
	 */
	@Override
	public void notifyChanged(final Notification notification) {
		final int featureId = notification.getFeatureID(RonmodelPackage.class);
		switch (featureId) {
		case RonmodelPackage.ON_NODE__X:
		case RonmodelPackage.ON_NODE__Y:
		case RonmodelPackage.ON_NODE__HEIGHT:
		case RonmodelPackage.ON_NODE__WIDTH:
		case RonmodelPackage.ON_NODE__NAME:
			refreshVisuals();
			break;
		case RonmodelPackage.ON_NODE__OUTGOING:
			refreshSourceConnections();
			refreshVisuals();
			break;
		case RonmodelPackage.ON_NODE__INCOMING:
			refreshTargetConnections();
			refreshVisuals();
			break;
		case RonmodelPackage.ON_NODE__COLOR:
			// a new color (must not be null!) has been set that the figure
			// should display
			Assert.isNotNull(notification.getNewValue(),
					"ONNodeEdit color has been set to null!");
			refreshVisuals();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see muvitorkit.gef.directedit.IDirectEditPart#getDirectEditFeatureID()
	 */
	@Override
	public int getDirectEditFeatureID() {
		return RonmodelPackage.ON_NODE__NAME;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see muvitorkit.gef.directedit.IDirectEditPart#getDirectEditValidator()
	 */
	@Override
	public ICellEditorValidator getDirectEditValidator() {
		// return new ICellEditorValidator() {
		// @Override
		// public String isValid(final Object value) {
		// if (value instanceof String) {
		// for (final ONNode node : getCastedModel().getObjectNet()
		// .getNodes()) {
		// if (node.getName().equals(value)
		// && node != getCastedModel()) {
		// return "There is already a node with name '"
		// + value + "'!";
		// }
		// }
		// }
		// return null;
		// }
		// };
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see muvitorkit.gef.directedit.IDirectEditPart#getValueLabelTextBounds()
	 */
	@Override
	public Rectangle getValueLabelTextBounds() {
		return getNodeFigure().getValueLabelTextBounds();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * muvitorkit.gef.directedit.IDirectEditPart#updateValueDisplay(java.lang
	 * .String)
	 */
	@Override
	public void updateValueDisplay(final String value) {
		getNodeFigure().updateValueDisplay(value);
	}

}
