package roneditor.model;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import muvitorkit.ui.MuvitorTreeEditor;
import muvitorkit.ui.utils.MuvitorNotifierService;
import muvitorkit.ui.utils.MuvitorNotifierService.IChangeListenerJob;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.gef.commands.Command;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.ui.progress.UIJob;

import roneditor.gef.actions.ron.RunAutoSimulationAction;
import roneditor.gef.commands.ron.HLTransitionApplyRuleCommand;
import roneditor.gef.commands.ron.HLTransitionFireCommand;
import roneditor.gef.commands.ron.HLTransitionSplitCommand;
import roneditor.gef.commands.ron.HLTransitionStandardCommand;
import ronmodel.HLArc;
import ronmodel.HLNode;
import ronmodel.HLPlace;
import ronmodel.HLPlaceType;
import ronmodel.HLTransition;
import ronmodel.HLTransitionKind;
import ronmodel.ONArc;
import ronmodel.ONNode;
import ronmodel.ONPlace;
import ronmodel.ONTransition;
import ronmodel.ObjectNet;
import ronmodel.RON;
import ronmodel.RonmodelPackage;
import ronmodel.Rule;
import ronmodel.Token;
import aggSim.Grammar;

/**
 * Provides some model specific tools for {@link HLTransition} such as checks
 * for validity and if the {@link HLTransition} can fire.
 * 
 * @author Tony Modica
 */

public class HLTransitionUtil {

	/**
	 * Install an adapter on the MuvitorNotifierService checking for all
	 * possible events that necessitate invalidation of HLTransition states.
	 */
	static {
		MuvitorNotifierService.addListener(new IChangeListenerJob() {
			@Override
			public void notifyChange(final EObject model,
					final Notification notification) {
				final int featureId = notification
						.getFeatureID(RonmodelPackage.class);
				/*
				 * case for RON not needed, because creation of nodes does not
				 * affect transitions, and modification of arcs is noticed by
				 * nodes.
				 */
				if (model instanceof HLTransition) {
					switch (featureId) {
					case RonmodelPackage.HL_TRANSITION__INCOMING:
					case RonmodelPackage.HL_TRANSITION__OUTGOING:
						invalidate((HLTransition) model, false);
						break;
					}
				} else if (model instanceof HLPlace) {
					switch (featureId) {
					case RonmodelPackage.HL_PLACE__TOKENS:
						invalidate((HLPlace) model, true);
						break;
					}
				} else if (model instanceof Rule) {
					switch (featureId) {
					case RonmodelPackage.RULE__NACS:
						invalidate(((Rule) model).getPlace(), true);
						break;
					}
				} else if (model instanceof ObjectNet) {
					switch (featureId) {
					case RonmodelPackage.OBJECT_NET__NODES:
					case RonmodelPackage.OBJECT_NET__ARCS:
					case RonmodelPackage.OBJECT_NET__MAPPINGS:
						invalidate((ObjectNet) model, true);
						break;
					}
				} else if (model instanceof ONPlace) {
					switch (featureId) {
					case RonmodelPackage.ON_PLACE__TOKENS:
					case RonmodelPackage.ON_PLACE__NAME:
						invalidate(((ONPlace) model).getObjectNet(), true);
						break;
					}
				} else if (model instanceof ONArc) {
					switch (featureId) {
					case RonmodelPackage.ON_ARC__WEIGHT:
						invalidate(((ONArc) model).getObjectNet(), true);
						break;
					}
				}
			}
		}, null);
	}

	private HLTransitionUtil() {
	}

	/**
	 * Updates the {@link HLTransition}'s enablement state attribute, which tell
	 * whether the transition is valid and can fire. The transition's
	 * "isEnabled" attribute will only be recalculated and changed if it has
	 * been unset.
	 * 
	 * @param transition
	 *            The transition to check
	 * @return true if the {@link HLTransition} can fire.
	 */
	synchronized public static void updateEnablement(
			final HLTransition transition) {
		// only proceed for valid transitions
		if (!validate(transition)) {
			transition.setActivated(false);
		}
		if (!transition.isSetActivated()) {
			switch (transition.getKind().getValue()) {
			case HLTransitionKind.APPLYRULE:
				transition.setActivated(canFireApplyRule(transition));
				break;
			case HLTransitionKind.FIRE:
				transition.setActivated(canFireFire(transition));
				break;
			case HLTransitionKind.SPLIT:
				transition.setActivated(canFireSplit(transition));
				break;
			case HLTransitionKind.STANDARD:
				transition.setActivated(canFireStandard(transition));
				break;
			}
		}
	}

	/**
	 * Schedules a check for validation and enablement of a transition. Will not
	 * schedule a checking job for a transition that already has a job
	 * scheduled.
	 * 
	 * @param transition
	 *            The transition to validate and check for enablement
	 * @return
	 */
	static void invalidate(final HLTransition transition,
			final boolean onlyEnablement) {
		if (transition == null) {
			return;
		}

		// Special for RunAutoSimulationAction: When simulating, we have to
		// update immediately and can not wait for
		// jobs without deadlocks! For performance, store validated transitions
		// till next step.
		if (RunAutoSimulationAction.isRunning()) {
			transition.unsetActivated();
			updateEnablement(transition);
			return;
		}

		// schedule at most one validating job for each transition at the same
		// time!
		if (Job.getJobManager().find(transition).length == 0) {
			if (!onlyEnablement) {
				transition.unsetValid();
			}
			transition.unsetActivated();
			final UIJob checkJob = new UIJob("Check '" + transition.getName()
					+ "'") {
				@Override
				public IStatus runInUIThread(final IProgressMonitor monitor) {
					monitor.setTaskName("Checking transition '"
							+ transition.getName() + "'");
					updateEnablement(transition);
					monitor.done();
					return Status.OK_STATUS;
				}

				@Override
				public boolean belongsTo(final Object family) {
					return family == transition;
				}
			};
			checkJob.setPriority(Job.BUILD);
			checkJob.schedule();
		}
	}

	public static void invalidate(final RON ron) {
		for (final HLNode node : ron.getNodes()) {
			if (node instanceof HLTransition) {
				final HLTransition trans = (HLTransition) node;
				HLTransitionUtil.invalidate((HLTransition) node,
						trans.isSetValid());
			}
		}
	}

	static void invalidate(final ObjectNet objectNet,
			final boolean onlyEnablement) {
		if (objectNet == null) {
			return;
		}
		final Rule rule = RuleUtil.getContainingRule(objectNet);
		if (rule != null) {
			invalidate(rule.getPlace(), onlyEnablement);
		} else {
			invalidate(objectNet.getPlace(), onlyEnablement);
		}
	}

	/**
	 * @param place
	 */
	static void invalidate(final HLPlace place, final boolean onlyEnablement) {
		if (place == null) {
			return;
		}
		for (final HLTransition transition : getEnvironment(place)) {
			invalidate(transition, onlyEnablement);
		}
	}

	/**
	 * Check if the transition is valid, hence if its pre- and postconditions
	 * hold (this is type dependent). The check will only be performed if the
	 * transition's "isValid" attribute has been unset. Otherwise, simply the
	 * attribute value will be returned.
	 * 
	 * @param transition
	 *            The {@link HLTransition} to check
	 * @return true if the transition is valid.
	 */
	private static boolean validate(final HLTransition transition) {
		if (!transition.isSetValid()) {
			switch (transition.getKind().getValue()) {
			case HLTransitionKind.APPLYRULE:
				transition.setValid(isValidApplyRule(transition));
				break;
			case HLTransitionKind.FIRE:
				transition.setValid(isValidFire(transition));
				break;
			case HLTransitionKind.SPLIT:
				transition.setValid(isValidSplit(transition));
				break;
			case HLTransitionKind.STANDARD:
				transition.setValid(isValidStandard(transition));
				break;
			}
		}
		return transition.isValid();
	}

	/**
	 * Return a {@link List} of all incoming {@link HLPlace} of the given
	 * {@link HLTransition} with {@link HLTransitionKind} <i>NET</i>.
	 * 
	 * @param transition
	 * @return the incoming places of type NET
	 */
	public static List<HLPlace> getIncomingNetPlaces(
			final HLTransition transition) {
		return getAdjacentPlaces(transition, HLPlaceType.NET, true);
	}

	/**
	 * Return a {@link List} of all incoming {@link HLPlace} of the given
	 * {@link HLTransition} with {@link HLTransitionKind} <i>RULE</i>.
	 * 
	 * @param transition
	 * @return the incoming places of type RULE
	 */
	public static List<HLPlace> getIncomingRulePlaces(
			final HLTransition transition) {
		return getAdjacentPlaces(transition, HLPlaceType.RULE, true);
	}

	/**
	 * Return a {@link List} of all outgoing {@link HLPlace} of the given
	 * {@link HLTransition} with {@link HLTransitionKind} <i>NET</i>.
	 * 
	 * @param transition
	 * @return the outgoing places of type NET
	 */
	public static List<HLPlace> getOutgoingNetPlaces(
			final HLTransition transition) {
		return getAdjacentPlaces(transition, HLPlaceType.NET, false);
	}

	/**
	 * Return a {@link List} of all outgoing {@link HLPlace} of the given
	 * {@link HLTransition} with {@link HLTransitionKind} <i>RULE</i>.
	 * 
	 * @param transition
	 * @return the outgoing places of type RULE
	 */
	public static List<HLPlace> getOutgoingRulePlaces(
			final HLTransition transition) {
		return getAdjacentPlaces(transition, HLPlaceType.RULE, false);
	}

	/**
	 * Return a {@link List} of all {@link HLPlace} of the given
	 * {@link HLTransition} with the given {@link HLTransitionKind}, according
	 * to isIncoming in pre or post-domain.
	 * 
	 * @param transition
	 * @param placeType
	 * @param isIncoming
	 * @return
	 */
	private static List<HLPlace> getAdjacentPlaces(
			final HLTransition transition, final int placeType,
			final boolean isIncoming) {
		final List<HLArc> adjacentArcs;
		if (isIncoming) {
			adjacentArcs = transition.getIncoming();
		} else {
			adjacentArcs = transition.getOutgoing();
		}
		final List<HLPlace> adjacentPlaces = new ArrayList<HLPlace>();
		for (final HLArc arc : adjacentArcs) {
			HLPlace adjacentPlace;
			if (isIncoming) {
				adjacentPlace = (HLPlace) arc.getSource();
			} else {
				adjacentPlace = (HLPlace) arc.getTarget();
			}
			if (adjacentPlace != null
					&& adjacentPlace.getType().getValue() == placeType) {
				adjacentPlaces.add(adjacentPlace);
			}
		}
		return adjacentPlaces;
	}

	static public Set<HLTransition> getEnvironment(final HLPlace place) {
		final Set<HLTransition> transitions = new HashSet<HLTransition>();
		for (final HLArc arc : place.getIncoming()) {
			transitions.add((HLTransition) arc.getSource());
		}
		for (final HLArc arc : place.getOutgoing()) {
			transitions.add((HLTransition) arc.getTarget());
		}
		return transitions;
	}

	/**
	 * Collect all {@link Token} from the incoming <i>NET</i> places.
	 * 
	 * @param transition
	 *            The {@link HLTransition} to check
	 * @return the tokens from the incoming net places
	 */
	public static List<ObjectNet> collectSourceNetTokens(
			final HLTransition transition) {
		final List<ObjectNet> tokens = new ArrayList<ObjectNet>();
		for (final HLPlace p : getIncomingNetPlaces(transition)) {
			// We dont allow multiple arcs, so we just take the first token
			if (!p.getTokens().isEmpty()) {
				tokens.add((ObjectNet) p.getTokens().get(0));
			}
		}
		return tokens;
	}

	/**
	 * Close all views showing an {@link ObjectNet} that would be used and not
	 * laid back on its originating HLPlace when firing the given transition.
	 * 
	 * @param transition
	 *            The {@link HLTransition} to fire
	 */
	public static void closeViewsForConsumedObjectNets(
			final HLTransition transition) {
		for (final ObjectNet objectnet : collectSourceNetTokens(transition)) {
			MuvitorTreeEditor.closeViewShowing(objectnet);
		}
	}

	/**
	 * Check if the transition of type standard can fire. Assuming valid
	 * transition: at least 1 pre place of any kind; if there is a NET post
	 * place there has to be at least 1 NET pre place, analogously for RULE
	 * places.
	 * 
	 * @param transition
	 *            The {@link HLTransition} to check
	 * @return true if there is at least one token on every incoming place and .
	 */
	private static boolean canFireStandard(final HLTransition transition) {
		return allIncomingPlacesNotEmpty(transition);
	}

	private static boolean allIncomingPlacesNotEmpty(
			final HLTransition transition) {
		for (final HLArc arc : transition.getIncoming()) {
			final HLPlace place = (HLPlace) arc.getSource();
			// no token on place? -> return
			if (place == null || place.getTokens().isEmpty()) {
				return false;
			}
		}
		// are there some tokens to fire at all?
		return !transition.getIncoming().isEmpty();
	}

	/**
	 * Check if the transition of type split can fire. Assuming valid
	 * transition: 1 incoming and 1 outgoing NET place, no RULE pre or post
	 * places
	 * 
	 * @return true if there is at least one token on every incoming place.
	 */
	private static boolean canFireSplit(final HLTransition transition) {
		return allIncomingPlacesNotEmpty(transition);
	}

	/**
	 * Check if the transition of type <i>APPLYRULE</i> can fire Assuming valid
	 * transition: 1 RULE pre place, at least 1 NET pre place
	 * 
	 * @param transition
	 *            The transition to check
	 * @return true if there is at least one token on every incoming place.
	 */
	private static boolean canFireApplyRule(final HLTransition transition) {
		if (!allIncomingPlacesNotEmpty(transition)) {
			return false;
		}
		// simulate rule application
		final List<ObjectNet> tokens = HLTransitionUtil
				.collectSourceNetTokens(transition);
		final ObjectNet unionObjectNet = ObjectNetUtil
				.unionIntoObjectNet(tokens);

		// Collect incoming NET/RULE HLPlaces
		final Rule rule = (Rule) getIncomingRulePlaces(transition).get(0)
				.getTokens().get(0);
		return new Grammar().isApplicable(rule, unionObjectNet);
	}

	/**
	 * Check if the transition of type fire can fire. Assuming valid transition:
	 * 1 NET pre and 1 NET post place, no RULE post places
	 * 
	 * @param transition
	 *            The {@link HLTransition} to check
	 * @return true if there is at least one token on every incoming place and
	 *         the {@link ONTransition} can actually fire.
	 */
	private static boolean canFireFire(final HLTransition transition) {
		if (!allIncomingPlacesNotEmpty(transition)) {
			return false;
		}
		final HLPlace p = (HLPlace) transition.getIncoming().get(0).getSource();
		for (final Object net : p.getTokens()) {
			for (final ONNode on : ((ObjectNet) net).getNodes()) {
				if (on instanceof ONPlace) {
					continue;
				}
				if (ONTransitionUtil.canFire((ONTransition) on)) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Check if the {@link HLTransition} of the type <i>STANDARD</i> is valid .
	 * Conditions: at least 1 pre place of any kind; if there is a NET post
	 * place there has to be at least 1 NET pre place, analogously for RULE
	 * places.
	 * 
	 * @param transition
	 *            The {@link HLTransition} to check
	 * @return true if the {@link HLTransition} is valid
	 */
	private static boolean isValidStandard(final HLTransition transition) {
		if (transition.getIncoming().isEmpty()) {
			return false;
		}
		if (!getOutgoingNetPlaces(transition).isEmpty()
				&& getIncomingNetPlaces(transition).isEmpty()) {
			return false;
		}
		if (!getOutgoingRulePlaces(transition).isEmpty()
				&& getIncomingRulePlaces(transition).isEmpty()) {
			return false;
		}
		return true;
	}

	/**
	 * Check if the {@link HLTransition} of the type <i>APPLYRULE</i> is valid.
	 * Condition: 1 RULE pre place, at least 1 NET pre place
	 * 
	 * @param transition
	 *            The {@link HLTransition} to check
	 * @return true if the {@link HLTransition} is valid
	 */
	private static boolean isValidApplyRule(final HLTransition transition) {
		if (getIncomingRulePlaces(transition).size() != 1) {
			return false;
		}
		if (getIncomingNetPlaces(transition).isEmpty()) {
			return false;
		}
		return true;
	}

	/**
	 * Check if the {@link HLTransition} of the type <i>FIRE</i> is valid.
	 * Conditions: 1 NET pre place, no RULE post places
	 * 
	 * @param transition
	 *            The {@link HLTransition} to check
	 * @return true if the {@link HLTransition} is valid
	 */
	private static boolean isValidFire(final HLTransition transition) {
		if (getIncomingNetPlaces(transition).size() != 1) {
			return false;
		}
		if (getOutgoingNetPlaces(transition).size() != 1) {
			return false;
		}
		if (!getIncomingRulePlaces(transition).isEmpty()) {
			return false;
		}
		if (!getOutgoingRulePlaces(transition).isEmpty()) {
			return false;
		}
		return true;
	}

	/**
	 * Check if the {@link HLTransition} of the type <i>SPLIT</i> is valid .
	 * Conditions: 1 incoming and 1 outgoing NET place, no RULE pre or post
	 * places
	 * 
	 * @param transition
	 *            The {@link HLTransition} to check
	 * @return true if the {@link HLTransition} is valid
	 */
	private static boolean isValidSplit(final HLTransition transition) {
		if (getIncomingNetPlaces(transition).size() != 1) {
			return false;
		}
		if (getOutgoingNetPlaces(transition).size() != 1) {
			return false;
		}
		if (!getIncomingRulePlaces(transition).isEmpty()) {
			return false;
		}
		if (!getOutgoingRulePlaces(transition).isEmpty()) {
			return false;
		}
		return true;
	}

	public static Command createFireCommand(final HLTransition transition) {
		if (!transition.isActivated()) {
			return null;
		}
		switch (transition.getKind().getValue()) {
		case HLTransitionKind.FIRE:
			return new HLTransitionFireCommand(transition, false);
		case HLTransitionKind.APPLYRULE:
			final HLTransitionApplyRuleCommand applyCommand = new HLTransitionApplyRuleCommand(
					transition);
			if (!applyCommand.canExecute()) {
				final Status status = new Status(IStatus.ERROR, "RONEditor", 0,
						"Could not apply the rule.", null);
				ErrorDialog.openError(null, "APPLYRULE",
						"Could not apply the rule", status);
			}
			return applyCommand;
		case HLTransitionKind.SPLIT:
			return new HLTransitionSplitCommand(transition);
		case HLTransitionKind.STANDARD:
			return new HLTransitionStandardCommand(transition);
		default:
			return null;
		}
	}
}
