package roneditor.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;

import ronmodel.ONArc;
import ronmodel.ONNode;
import ronmodel.ObjectNet;
import ronmodel.RonmodelFactory;
import ronmodel.Token;

public class ObjectNetUtil {

	private ObjectNetUtil() {
	}

	/**
	 * Traverse the tree recursively and tag all unvisited nodes in the given
	 * {@link Collection}.
	 * 
	 * @param list
	 *            {@link Collection} of tagged {@link ONNode}s
	 * @param node
	 *            {@link ONNode} to traverse.
	 */
	private static void visitNode(final Collection<ONNode> list,
			final ONNode node) {
		list.add(node);
		for (final ONArc arc : node.getOutgoing()) {
			final ONNode target = arc.getTarget();
			if (!list.contains(target)) {
				visitNode(list, target);
			}
		}
		for (final ONArc arc : node.getIncoming()) {
			final ONNode source = arc.getSource();
			if (!list.contains(source)) {
				visitNode(list, source);
			}
		}
	}

	/**
	 * Take the first node from an {@link ObjectNet} and build it's spanning
	 * tree, then return the connected component derived from the spanning tree
	 * to a new {@link ObjectNet} and return it.
	 * 
	 * @param on
	 *            {@link ObjectNet} to split off the first connected component
	 * @return The new {@link ObjectNet} with the first connected component
	 */
	private static ObjectNet splitFirst(final ObjectNet on) {
		final ObjectNet result = RonmodelFactory.eINSTANCE.createObjectNet();
		final EList<ONNode> nodes = on.getNodes();
		final ONNode n0 = nodes.get(0);
		final HashSet<ONNode> list = new HashSet<ONNode>();
		visitNode(list, n0);
		moveConnectedComponent(list, result);
		result.setName(on.getName());
		return result;
	}

	/**
	 * Performs an very lazy split operation on an {@link ObjectNet}, hence
	 * splits all connected components into a {@link List} of {@link ObjectNet}.
	 * 
	 * @param on
	 *            the objectnet to be split
	 * @return objectnets being the components of the passed object net
	 */
	public static Collection<ObjectNet> split(final ObjectNet on) {
		final List<ObjectNet> result = new ArrayList<ObjectNet>();
		if (on.getNodes() == null) {
			return result;
		}
		while (on.getNodes().size() > 0) {
			result.add(splitFirst(on));
		}
		return result;
	}

	/**
	 * Move a connected component to the target {@link ObjectNet}, hence move
	 * all the {@link ONNode} in the given {@link Collection} with it's outgoing
	 * {@link ONArc} to the target {@link ObjectNet}.
	 * 
	 * @param nodes
	 * @param target
	 */
	public static void moveConnectedComponent(final Collection<ONNode> nodes,
			final ObjectNet target) {
		// Copy as we change the list while iterating
		final List<ONNode> nodesCopy = new ArrayList<ONNode>(nodes);
		for (final ONNode node : nodesCopy) {
			node.setObjectNet(target);
			final List<ONArc> outgoing = new ArrayList<ONArc>(
					node.getOutgoing());
			for (final ONArc arc : outgoing) {
				arc.setObjectNet(target);
			}
		}

	}

	public static ObjectNet unionIntoObjectNet(final List<ObjectNet> tokens) {
		final StringBuilder targetName = new StringBuilder();
		final Collection<EObject> allTokenContent = new ArrayList<EObject>();
		for (final Token net : tokens) {
			if (allTokenContent.isEmpty()) {
				allTokenContent.addAll(EcoreUtil.copyAll(net.eContents()));
				targetName.append(net.getName());
			} else {
				allTokenContent.addAll(EcoreUtil.copyAll(net.eContents()));
				targetName.append("+");
				targetName.append(net.getName());
			}
		}
		final ObjectNet target = RonmodelFactory.eINSTANCE.createObjectNet();
		target.setName(targetName.toString());
		target.setPlace(null);
		for (final EObject object : allTokenContent) {
			if (object instanceof ONNode) {
				((ONNode) object).setObjectNet(target);
			} else if (object instanceof ONArc) {
				((ONArc) object).setObjectNet(target);
			}
		}
		return target;
	}
}
