package roneditor.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import muvitorkit.ui.IDUtil;
import muvitorkit.ui.MuvitorTreeEditor;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.emf.ecore.EObject;

import ronmodel.Mapping;
import ronmodel.NAC;
import ronmodel.ONArc;
import ronmodel.ONNode;
import ronmodel.ONPlace;
import ronmodel.ONTransition;
import ronmodel.Rule;

public class RuleChecker {

	private static final String CAUSING_RULE_ATTR = "CausingRule";

	private RuleChecker() {
	}

	/**
	 * Check if the mappings are compatible, hence if the source and target is
	 * of the same type
	 */
	private static void checkCompatibility(final Rule rule,
			final List<Mapping> mappings) {
		for (final Mapping mapping : mappings) {
			final ONNode source = mapping.getSource();
			final ONNode target = mapping.getTarget();

			if (source.getClass() != target.getClass()) {
				createErrorMarker(source, rule.getName(),
						"Mapping is not compatible (" + source.getClass()
								+ "->" + target.getClass() + ")");
				// this is a severe error, skip the next tests for this mapping
				continue;
			}
			// seb: are source and target name the same?
			// ( if sourcename not empty folie 09RuleEditor.pdf page 7 )
			// we treat " " and "" as compatible names to avoid a possible EMF
			// bug
			final String sourceName = source.getName().trim();
			final String targetName = target.getName().trim();
			if (targetName.compareTo(sourceName) != 0) {
				// if in LHS empty and target in NAC it is ok
				if (!(target.getObjectNet() instanceof NAC)
						|| !sourceName.isEmpty()) {

					if ((!sourceName.equals(" ") || !targetName.isEmpty())
							&& (!sourceName.isEmpty() || !targetName
									.equals(" "))) {
						createErrorMarker(source, rule.getName(),
								"Names are not compatible (" + sourceName + "@"
										+ source.getObjectNet().getName()
										+ " -> " + targetName + "@"
										+ target.getObjectNet().getName() + ")");
					}
				}
			}

			// check capacities and token numbers (rule morphisms have to be
			// strict!)
			if (target instanceof ONPlace) {
				if (((ONPlace) target).getCapacity() != ((ONPlace) source)
						.getCapacity()) {
					createErrorMarker(
							source,
							rule.getName(),
							"Capacities are not compatible ("
									+ source.getName() + "->"
									+ target.getName() + ")");
				}
				if (((ONPlace) source).getTokens() != ((ONPlace) target)
						.getTokens()) {
					createErrorMarker(source, rule.getName(),
							"Mapping is not strict (different amounts of token in "
									+ sourceName + "@"
									+ source.getObjectNet().getName() + " -> "
									+ targetName + "@"
									+ target.getObjectNet().getName() + ")");
				}
			}
		}
	}

	private static void createErrorMarker(final EObject model,
			final String location, final String message) {
		final IMarker newMarker = IDUtil.getHostEditor(model)
				.createErrorMarker(model, location, message);
		try {
			newMarker.setAttribute(CAUSING_RULE_ATTR, IDUtil
					.getRealURIFragment(RuleUtil.getContainingRule(model)));
		} catch (final CoreException e) {
		}
	}

	// /**
	// * Helper function for checking injectivity, performs actual checks on the
	// * mapping for one-to-one relation.
	// *
	// * @param mappings
	// * of mappings
	// */
	// private static void checkMappingsForInjectivity(final Rule rule,
	// final List<Mapping> mappings) {
	// final HashMap<ONNode, Mapping> targets = new HashMap<ONNode, Mapping>();
	// final HashMap<ONNode, Mapping> sources = new HashMap<ONNode, Mapping>();
	// for (final Mapping mapping : mappings) {
	// final ONNode source = mapping.getSource();
	// final ONNode target = mapping.getTarget();
	// final Mapping sourceMapping = sources.get(target);
	// if (sourceMapping != null) {
	// createErrorMarker(source, rule.getName(),
	// "There is already a mapping "
	// + sourceMapping.getSource().getName() + "->"
	// + sourceMapping.getTarget().getName());
	// }
	// final Mapping targetMapping = targets.get(target);
	// if (targetMapping != null) {
	// createErrorMarker(source, rule.getName(),
	// "There is already a mapping "
	// + targetMapping.getSource().getName() + "->"
	// + targetMapping.getTarget().getName());
	// }
	// sources.put(source, mapping);
	// targets.put(target, mapping);
	// }
	// }

	/**
	 * Check if pre and post conditions hold. Formal: f_P^{\oplus} \circ pre_1 =
	 * pre_2 \circ f_t f_P^{\oplus} \circ post_1 = post_2 \circ f_t
	 */
	private static void checkPreAndPostConditions(final Rule rule,
			final List<Mapping> mappings) {
		// prepare helper lists and maps
		final HashMap<ONPlace, ONPlace> placeMap = new HashMap<ONPlace, ONPlace>();
		final HashMap<ONTransition, ONTransition> transMap = new HashMap<ONTransition, ONTransition>();
		final List<ONTransition> mappedLhsTransitions = new ArrayList<ONTransition>();
		for (final Mapping mapping : mappings) {
			final ONNode node = mapping.getSource();
			if (node instanceof ONTransition) {
				final ONTransition trans = (ONTransition) node;
				transMap.put(trans, (ONTransition) mapping.getTarget());
				mappedLhsTransitions.add(trans);
			} else {
				placeMap.put((ONPlace) node, (ONPlace) mapping.getTarget());
			}
		}

		// check pre-conditions
		final HashMap<ONNode, ONArc> mappedPlaceToArcMap = new HashMap<ONNode, ONArc>();
		for (final ONTransition lhsTrans : mappedLhsTransitions) {
			// prepare location string for error marker
			final String transLocation = lhsTrans.getName() + "@"
					+ lhsTrans.getObjectNet().getName() + " [" + rule.getName()
					+ "]";
			final ONTransition mappedTrans = transMap.get(lhsTrans);
			// for pre-check
			mappedPlaceToArcMap.clear();
			// remember places on incoming arcs
			for (final ONArc arc : mappedTrans.getIncoming()) {
				mappedPlaceToArcMap.put(arc.getSource(), arc);
			}
			// test all incoming arcs
			for (final ONArc arcI : lhsTrans.getIncoming()) {

				final ONNode node = placeMap.get(arcI.getSource());
				if (node == null) {
					createErrorMarker(arcI, transLocation, "Unmapped Place '"
							+ arcI.getTarget().getName()
							+ "' is in Pre-Condition of mapped Transition '"
							+ lhsTrans.getName() + "' ("
							+ lhsTrans.getObjectNet().getName() + ")");
					continue;
				}
				final ONArc arc = mappedPlaceToArcMap.get(node);
				if (arc == null) {
					createErrorMarker(arcI, transLocation,
							"No corresponding Arc from Place "
									+ arcI.getSource().getName()
									+ " found in Pre-Condition ("
									+ node.getObjectNet().getName() + ")");
					continue;
				}
				if (arc.getWeight() != arc.getWeight()) {
					createErrorMarker(arc, transLocation,
							"Different weights for Arcs from Place "
									+ arc.getSource().getName() + " and Place "
									+ arc.getSource().getName()
									+ " in Post-Condition ("
									+ node.getObjectNet().getName() + ")");
				}
				mappedPlaceToArcMap.remove(node);
			}
			if (!mappedPlaceToArcMap.isEmpty()) {
				createErrorMarker(lhsTrans, transLocation,
						"Missing Arc(s) in Pre-Conditions (lhs)");
			}

			// for post-check
			mappedPlaceToArcMap.clear();
			for (final ONArc i : mappedTrans.getOutgoing()) {
				mappedPlaceToArcMap.put(i.getTarget(), i);
			}
			for (final ONArc arcI : lhsTrans.getOutgoing()) {
				final ONNode node = placeMap.get(arcI.getTarget());
				if (node == null) {
					createErrorMarker(arcI, transLocation, "Unmapped Place '"
							+ arcI.getTarget().getName()
							+ "' is in Post-Condition of mapped Transition '"
							+ lhsTrans.getName() + "' ("
							+ lhsTrans.getObjectNet().getName() + ")");
					continue;
				}
				final ONArc arc = mappedPlaceToArcMap.get(node);
				if (arc == null) {
					createErrorMarker(arcI, transLocation,
							"No corresponding Arc to Place "
									+ arcI.getTarget().getName()
									+ " found in Post-Condition ("
									+ node.getObjectNet().getName() + ")");
					continue;
				}
				if (arc.getWeight() != arcI.getWeight()) {
					createErrorMarker(arcI, transLocation,
							"Different weights for Arcs to Place "
									+ arcI.getTarget().getName()
									+ " and Place " + arc.getTarget().getName()
									+ " in Post-Condition ("
									+ node.getObjectNet().getName() + ")");
				}
				mappedPlaceToArcMap.remove(node);
			}
			if (!mappedPlaceToArcMap.isEmpty()) {
				createErrorMarker(lhsTrans, transLocation,
						"Missing Arc(s) in Post-Condition (lhs)");
			}
		}
	}

	public static boolean isRuleOk(final Rule rule) {
		final String ruleURIString = IDUtil.getRealURIFragment(rule);
		final MuvitorTreeEditor editor = IDUtil.getHostEditor(rule);
		for (final IMarker marker : editor.findProblemMarkers(
				CAUSING_RULE_ATTR, ruleURIString)) {
			try {
				marker.delete();
			} catch (final CoreException e1) {
				e1.printStackTrace();
			}
		}

		final List<Mapping> mappings = RuleUtil.getAllCompleteMappings(rule);
		checkCompatibility(rule, mappings);
		// checkMappingsForInjectivity(rule, mappings);
		checkPreAndPostConditions(rule, mappings);

		final ArrayList<IMarker> newFoundProblems = editor.findProblemMarkers(
				CAUSING_RULE_ATTR, ruleURIString);
		return newFoundProblems.isEmpty();
	}
}
